@@ -5,20 +5,28 @@
 */
 #include "output_model.h"
 
-#include "../common/utils.h"
-#include "kscreen/edid.h"
+#include <cstdint>
 
+#include <kscreen/edid.h>
+#include <kscreen/mode.h>
+
+#include "../common/utils.h"
 #include "config_handler.h"
 
 #include <KLocalizedString>
 
 #include <QRect>
+#include <numeric>
 
 OutputModel::OutputModel(ConfigHandler *configHandler)
     : QAbstractListModel(configHandler)
     , m_config(configHandler)
 {
-    connect(this, &OutputModel::dataChanged, this, &OutputModel::changed);
+    connect(m_config->config().data(), &KScreen::Config::prioritiesChanged, this, [this]() {
+        if (rowCount() > 0) {
+            Q_EMIT dataChanged(createIndex(0, 0), createIndex(rowCount() - 1, 0), {PriorityRole});
+        }
+    });
 }
 
 int OutputModel::rowCount(const QModelIndex &parent) const
@@ -53,8 +61,8 @@ QVariant OutputModel::data(const QModelIndex &index, int role) const
         return output->isEnabled();
     case InternalRole:
         return output->type() == KScreen::Output::Type::Panel;
-    case PrimaryRole:
-        return output->isPrimary();
+    case PriorityRole:
+        return output->priority();
     case SizeRole:
         return output->geometry().size();
     case PositionRole:
@@ -99,6 +107,8 @@ QVariant OutputModel::data(const QModelIndex &index, int role) const
         return static_cast<uint32_t>(output->vrrPolicy());
     case RgbRangeRole:
         return static_cast<uint32_t>(output->rgbRange());
+    case InteractiveMoveRole:
+        return m_outputs[index.row()].moving;
     }
     return QVariant();
 }
@@ -117,7 +127,6 @@ bool OutputModel::setData(const QModelIndex &index, const QVariant &value, int r
             if (output.pos == val) {
                 return false;
             }
-
             snap(output, val);
             m_outputs[index.row()].pos = val;
             updatePositions();
@@ -131,14 +140,13 @@ bool OutputModel::setData(const QModelIndex &index, const QVariant &value, int r
             return setEnabled(index.row(), value.toBool());
         }
         break;
-    case PrimaryRole:
-        if (value.canConvert<bool>()) {
-            bool primary = value.toBool();
-            if (output.ptr->isPrimary() == primary) {
+    case PriorityRole:
+        if (value.canConvert<uint32_t>()) {
+            const uint32_t priority = value.toUInt();
+            if (output.ptr->priority() == priority) {
                 return false;
             }
-            m_config->config()->setPrimaryOutput(output.ptr);
-            Q_EMIT dataChanged(index, index, {role});
+            m_config->config()->setOutputPriority(output.ptr, priority);
             return true;
         }
         break;
@@ -233,6 +241,14 @@ bool OutputModel::setData(const QModelIndex &index, const QVariant &value, int r
             Q_EMIT dataChanged(index, index, {role});
             return true;
         }
+        break;
+    case InteractiveMoveRole:
+        if (value.canConvert<bool>()) {
+            m_outputs[index.row()].moving = value.toBool();
+            Q_EMIT dataChanged(index, index, {role});
+            return true;
+        }
+        break;
     }
     return false;
 }
@@ -242,7 +258,7 @@ QHash<int, QByteArray> OutputModel::roleNames() const
     QHash<int, QByteArray> roles = QAbstractItemModel::roleNames();
     roles[EnabledRole] = "enabled";
     roles[InternalRole] = "internal";
-    roles[PrimaryRole] = "primary";
+    roles[PriorityRole] = "priority";
     roles[SizeRole] = "size";
     roles[PositionRole] = "position";
     roles[NormalizedPositionRole] = "normalizedPosition";
@@ -262,6 +278,7 @@ QHash<int, QByteArray> OutputModel::roleNames() const
     roles[OverscanRole] = "overscan";
     roles[VrrPolicyRole] = "vrrPolicy";
     roles[RgbRangeRole] = "rgbRange";
+    roles[InteractiveMoveRole] = "interactiveMove";
     return roles;
 }
 
@@ -290,9 +307,6 @@ void OutputModel::add(const KScreen::OutputPtr &output)
     }
     m_outputs.insert(i, Output(output, pos));
 
-    connect(output.data(), &KScreen::Output::isPrimaryChanged, this, [this, output]() {
-        roleChanged(output->id(), PrimaryRole);
-    });
     endInsertRows();
 
     connect(output.data(), &KScreen::Output::modesChanged, this, [this, output]() {
@@ -325,9 +339,9 @@ void OutputModel::remove(int outputId)
     }
 }
 
-void OutputModel::resetPosition(const Output &output)
+void OutputModel::resetPosition(Output &output)
 {
-    if (output.posReset.x() < 0) {
+    if (!output.posReset.has_value()) {
         // KCM was closed in between.
         for (const Output &out : qAsConst(m_outputs)) {
             if (out.ptr->id() == output.ptr->id()) {
@@ -338,10 +352,46 @@ void OutputModel::resetPosition(const Output &output)
             }
         }
     } else {
-        output.ptr->setPos(/*output.ptr->pos() - */ output.posReset);
+        QPoint reset = output.posReset.value();
+        output.posReset.reset();
+        QPoint shift = QPoint(0, 0);
+
+        if (reset.x() < 0) {
+            shift.setX(-reset.x());
+            reset.setX(0);
+        }
+        if (reset.y() < 0) {
+            shift.setY(-reset.y());
+            reset.setY(0);
+        }
+
+        for (Output &out : m_outputs) {
+            if (out.ptr->id() == output.ptr->id()) {
+                continue;
+            }
+            if (positionable(out)) {
+                out.ptr->setPos(out.ptr->pos() + shift);
+            }
+        }
+        output.ptr->setPos(reset);
     }
 }
 
+QPoint OutputModel::mostTopLeftLocationOfPositionableOutputOptionallyIgnoringOneOfThem(std::optional<KScreen::OutputPtr> ignored) const
+{
+    auto foldTopLeft = [=](std::optional<QPoint> a, const Output &out) {
+        if (!positionable(out) || (ignored.has_value() && out.ptr->id() == ignored.value()->id())) {
+            return a;
+        }
+        if (a.has_value()) {
+            return std::optional(QPoint(std::min(a.value().x(), out.pos.x()), std::min(a.value().y(), out.pos.y())));
+        } else {
+            return std::optional(out.pos);
+        }
+    };
+    return std::accumulate(m_outputs.constBegin(), m_outputs.constEnd(), std::optional<QPoint>(), foldTopLeft).value_or(QPoint(0, 0));
+}
+
 bool OutputModel::setEnabled(int outputIndex, bool enable)
 {
     Output &output = m_outputs[outputIndex];
@@ -356,11 +406,23 @@ bool OutputModel::setEnabled(int outputIndex, bool enable)
         resetPosition(output);
 
         setResolution(outputIndex, resolutionIndex(output.ptr));
-        reposition();
     } else {
-        output.posReset = output.ptr->pos();
+        // assuming it was already properly normalized, so current topleft (without disabling) is (0,0)
+        const QPoint topLeft = mostTopLeftLocationOfPositionableOutputOptionallyIgnoringOneOfThem(std::optional(output.ptr));
+
+        QPoint reset = output.ptr->pos();
+        if (topLeft.x() > 0) {
+            reset.setX(-topLeft.x());
+        }
+        if (topLeft.y() > 0) {
+            reset.setY(-topLeft.y());
+        }
+
+        output.posReset = std::optional(reset);
     }
 
+    reposition();
+
     QModelIndex index = createIndex(outputIndex, 0);
     Q_EMIT dataChanged(index, index, {EnabledRole});
     return true;
@@ -714,7 +776,7 @@ bool OutputModel::setReplicationSourceIndex(int outputIndex, int sourceIndex)
             return false;
         }
         m_config->setReplicationSource(output.ptr, source);
-        output.posReset = output.ptr->pos();
+        output.posReset = std::optional(output.ptr->pos());
         output.ptr->setPos(source->pos());
         output.ptr->setExplicitLogicalSize(m_config->config()->logicalSizeForOutput(*source));
     }
@@ -769,11 +831,6 @@ QVariantList OutputModel::replicasModel(const KScreen::OutputPtr &output) const
     return ret;
 }
 
-void OutputModel::roleChanged(int outputId, OutputRoles role)
-{
-    rolesChanged(outputId, {role});
-}
-
 void OutputModel::rolesChanged(int outputId, const QVector<int> &roles)
 {
     const auto index = indexForOutputId(outputId);
@@ -798,6 +855,11 @@ bool OutputModel::positionable(const Output &output) const
     return output.ptr->isPositionable();
 }
 
+bool OutputModel::isMoving() const
+{
+    return std::any_of(m_outputs.cbegin(), m_outputs.cend(), std::mem_fn(&Output::moving));
+}
+
 void OutputModel::reposition()
 {
     int x = 0;
@@ -839,39 +901,9 @@ void OutputModel::reposition()
     m_config->normalizeScreen();
 }
 
-QPoint OutputModel::originDelta() const
-{
-    int x = 0;
-    int y = 0;
-
-    // Find first valid output.
-    for (const auto &out : m_outputs) {
-        if (positionable(out)) {
-            x = out.pos.x();
-            y = out.pos.y();
-            break;
-        }
-    }
-
-    for (int i = 1; i < m_outputs.size(); i++) {
-        if (!positionable(m_outputs[i])) {
-            continue;
-        }
-        const QPoint &cmp = m_outputs[i].pos;
-
-        if (cmp.x() < x) {
-            x = cmp.x();
-        }
-        if (cmp.y() < y) {
-            y = cmp.y();
-        }
-    }
-    return QPoint(x, y);
-}
-
 void OutputModel::updatePositions()
 {
-    const QPoint delta = originDelta();
+    const QPoint delta = mostTopLeftLocationOfPositionableOutputOptionallyIgnoringOneOfThem();
     for (int i = 0; i < m_outputs.size(); i++) {
         const auto &out = m_outputs[i];
         if (!positionable(out)) {
@@ -884,44 +916,6 @@ void OutputModel::updatePositions()
             Q_EMIT dataChanged(index, index, {NormalizedPositionRole});
         }
     }
-    updateOrder();
-}
-
-void OutputModel::updateOrder()
-{
-    auto order = m_outputs;
-    std::sort(order.begin(), order.end(), [](const Output &a, const Output &b) {
-        const int xDiff = b.ptr->pos().x() - a.ptr->pos().x();
-        const int yDiff = b.ptr->pos().y() - a.ptr->pos().y();
-        if (xDiff > 0) {
-            return true;
-        }
-        if (xDiff == 0 && yDiff > 0) {
-            return true;
-        }
-        return false;
-    });
-
-    for (int i = 0; i < order.size(); i++) {
-        for (int j = 0; j < m_outputs.size(); j++) {
-            if (order[i].ptr->id() != m_outputs[j].ptr->id()) {
-                continue;
-            }
-            if (i != j) {
-                beginMoveRows(QModelIndex(), j, j, QModelIndex(), i);
-                m_outputs.remove(j);
-                m_outputs.insert(i, order[i]);
-                endMoveRows();
-            }
-            break;
-        }
-    }
-
-    // TODO: Could this be optimized by only outputs updating where replica indices changed?
-    for (int i = 0; i < m_outputs.size(); i++) {
-        QModelIndex index = createIndex(i, 0);
-        Q_EMIT dataChanged(index, index, {ReplicasModelRole});
-    }
 }
 
 bool OutputModel::normalizePositions()
@@ -946,7 +940,7 @@ bool OutputModel::normalizePositions()
 bool OutputModel::positionsNormalized() const
 {
     // There might be slight deviations because of snapping.
-    return originDelta().manhattanLength() < 5;
+    return mostTopLeftLocationOfPositionableOutputOptionallyIgnoringOneOfThem().manhattanLength() < 5;
 }
 
 const int s_snapArea = 80;
@@ -992,6 +986,17 @@ bool snapToLeft(const QRect &target, const QSize &size, QPoint &dest)
     return false;
 }
 
+bool snapHorizontal(const QRect &target, const QSize &size, QPoint &dest)
+{
+    if (snapToRight(target, size, dest)) {
+        return true;
+    }
+    if (snapToLeft(target, size, dest)) {
+        return true;
+    }
+    return false;
+}
+
 bool snapToMiddle(const QRect &target, const QSize &size, QPoint &dest)
 {
     const int outputMid = dest.y() + size.height() / 2;
@@ -1051,14 +1056,59 @@ bool snapVertical(const QRect &target, const QSize &size, QPoint &dest)
 void OutputModel::snap(const Output &output, QPoint &dest)
 {
     const QSize size = output.ptr->geometry().size();
-    for (const Output &out : qAsConst(m_outputs)) {
+    const QRect outputRect(dest, size);
+
+    QVector<std::reference_wrapper<const Output>> positionableOutputs;
+    positionableOutputs.reserve(m_outputs.size());
+    std::copy_if(m_outputs.cbegin(), m_outputs.cend(), std::back_inserter(positionableOutputs), [](const Output &output) {
+        return output.ptr->isPositionable();
+    });
+
+    // Special case for two outputs, we want to make sure they always touch;
+    if (positionableOutputs.size() == 2) {
+        const Output &other = positionableOutputs.at(0).get().ptr->id() == output.ptr->id() ? positionableOutputs.at(1) : positionableOutputs.at(0);
+        const QRect target(other.pos, other.ptr->geometry().size());
+        const bool xOverlap = dest.x() <= target.x() + target.width() && target.x() <= dest.x() + size.width();
+        const bool yOverlap = dest.y() <= target.y() + target.height() && target.y() <= dest.y() + size.height();
+        // Special special case, snap to center if centers are close
+        if (std::abs((outputRect.center() - target.center()).manhattanLength()) < s_snapArea * 2) {
+            dest = target.center() - (outputRect.center() - outputRect.topLeft());
+            return;
+        }
+        if (xOverlap) {
+            const int topDist = std::abs(dest.y() - target.y() - target.height());
+            const int bottomDist = std::abs(outputRect.y() + outputRect.height() - target.y());
+            if (topDist < bottomDist) {
+                dest.setY(target.y() + target.height());
+            } else {
+                dest.setY(target.y() - size.height());
+            }
+            // Secondary snap to align the other edges if close - right to right, left to left
+            snapHorizontal(target, size, dest);
+            return;
+        }
+        if (yOverlap) {
+            const int leftDist = std::abs(dest.x() - target.x() - target.width());
+            const int rightDiff = std::abs(outputRect.x() + outputRect.width() - target.x());
+            if (leftDist < rightDiff) {
+                dest.setX(target.x() + target.width());
+            } else {
+                dest.setX(target.x() - size.width());
+            }
+            // Secondary snap to align the other edges if close - top to top, bottom to bottom, center to center
+            snapVertical(target, size, dest);
+            return;
+        }
+        // No overlap at all can happen at a corner, do not let the output move away
+        dest = output.pos;
+        return;
+    }
+
+    for (const Output &out : qAsConst(positionableOutputs)) {
         if (out.ptr->id() == output.ptr->id()) {
             // Can not snap to itself.
             continue;
         }
-        if (!positionable(out)) {
-            continue;
-        }
 
         const QRect target(out.pos, out.ptr->geometry().size());
 
